
import os
import shutil

from .sage_types import pi, sin, cos, arccos, arcsin, arctan, sqrt, span, vector, Matrix, n

VERTEX_RADIUS = '.03'
AXIS_RADIUS = '.008'
CONE_RADIUS = '.03'
CONE_HEIGHT = '.09'
EDGE_RADIUS = '.005'

def fix_x3d_str(x3d_str):
    x3d_str = x3d_str.replace('\n','')
    x3d_split = x3d_str.split('>')
    x3d_tags = []
    for string in x3d_split:
        split_str = string.split('<')
        for ss in split_str:
            if ss.replace(' ','') != '':
                x3d_tags.append(ss)
    x3d_str = ''
    for tag in x3d_tags:
        if tag[-1] == '/':
            tag_name = tag.split(' ')[0]
            formatted_tag = '<'+tag[:-1]+'></{}>'.format(tag_name)
        else:
            formatted_tag = '<'+tag+'>'
        x3d_str += formatted_tag
    return x3d_str

def make_x3d_text(text,coords,scale,size,color):
    text = "\n<Transform translation='{0}' scale='{1}'><Billboard axisOfRotation='0 0 0'><Shape><Appearance><Material ambientIntensity='0.0933' diffuseColor='{4} {5} {6}' shininess='0.51' specularColor='{4} {5} {6}'></Material></Appearance><Text string='{2}' solid='false'><FontStyle family="'Times' 'Orbitron'" quality='3.0' size='{3}'></FontStyle></Text></Shape></Billboard></Transform>".format(coords,scale,text,size,color[0],color[1],color[2])
    return text

def make_x3d_label(genus,bdy,coords,scale,size1,size2,color):
    x=size1*.78
    y=size1*(-.3)
    text = "\n<Transform translation='{0} {1} {2}' scale='{3}'><Billboard axisOfRotation='0 0 0'><Group><Shape><Appearance><Material ambientIntensity='0.0933' diffuseColor='{9} {10} {11}' shininess='0.51' specularColor='{9} {10} {11}'></Material></Appearance><Text string='{6}' solid='false'><FontStyle family="'Times' 'Orbitron'" quality='3.0' size='{4}'></FontStyle></Text></Shape><Transform translation='{12} {13} 0.0'><Shape><Appearance><Material ambientIntensity='0.0933' diffuseColor='{9} {10} {11}' shininess='0.51' specularColor='{9} {10} {11}'></Material></Appearance><Text string='&#160 &#160 {7},{8}' solid='false'><FontStyle family="'Times' 'Orbitron'" quality='3.0' size='{5}'></FontStyle></Text></Shape></Transform></Group></Billboard></Transform>".format(coords[0],coords[1],coords[2],scale,size1,size2,'&#931',genus,bdy,color[0],color[1],color[2],x,y)
    return text

def make_x3d_vertex(coords,radius):
    vert = "\n<Transform translation='{}'><Shape>\n<Appearance>\n<Material ambientIntensity='0.0933' diffuseColor='1 1 1' shininess='0.51' specularColor='0.46 0.46 0.46'></Material>\n</Appearance>\n<Sphere ccw='true' lit='true' radius='{}' solid='true' subdivision='24,24'>\n</Sphere>\n</Shape>\n</Transform>".format(coords,radius)
    return vert

def make_x3d_edge(v1_coords, v2_coords,radius,color):
    v1 = v1_coords
    v2 = v2_coords
    str_v1 = '{} {} {}'.format(rnd(v1[0]),rnd(v1[1]),rnd(v1[2]))
    v = v2-v1
    rho = length = rnd(v.norm())
    trans = length/2
    phi = rnd(arccos(v[2]/rho))
    if abs(v[0]) < 0.0001:
        theta = n(pi/2) if v[1]>0 else n(-pi/2)
    else:
        theta = rnd(arctan(v[1]/v[0]))
        if v[0] < 0:
            theta = rnd(theta + pi)
    edge = "\n<Transform translation='{5}'>\n<Transform rotation='0 0 1 {4}' center='0 0 0'>\n<Transform rotation='0 1 0 {3}' center='0 0 0'>\n<Transform rotation='1 0 0 1.57079' center='0 0 0'>\n<Transform translation='0.0 {2} 0.0'>\n<Shape>\n<Appearance>\n<Material diffuseColor='{6} {7} {8}' shininess='1.0' specularColor='{6} {7} {8}'>\n</Material>\n</Appearance>\n<Cylinder radius='{0}' height='{1}' solid='true'>\n</Cylinder>\n</Shape>\n</Transform>\n</Transform>\n</Transform>\n</Transform>\n</Transform>".format(radius,length,trans,phi,theta,str_v1,color[0],color[1],color[2])
    return edge    

def rnd(x):
    if abs(x) < .00001:
        return 0
    else:
        return n(x,20)   

def rnd_vec(v):
    return vector(map(rnd,v))

#def scale_vec(v, factor):
#    return v*factor

#def make_x3d_schlegel(schlegel_projection, scale_factor, color):
#    SP = schlegel_projection
#    edges = SP.lines
#    coords = list(map(lambda v: v*scale_factor,list(map(rnd_vec,SP.coordinates_of(SP.points)))))
#    x3d_str = '\n'
#    x3d_str += "<Shape>\n<Appearance><Material ambientIntensity='0' emissiveColor='{2} {3} {4}' diffuseColor='{2} {3} {4}' specularColor='{2} {3} {4}' shininess='0.0078125' transparency='0'></Material></Appearance><IndexedLineSet coordIndex='{0}' colorPerVertex='false'>\n<Coordinate point='{1}'>\n</Coordinate>\n</IndexedLineSet>\n</Shape>"
#    coordIndex = ''
#    coordPoint = str(coords).replace('0.00000','0.0').replace('[(','').replace(')]','').replace(',','').replace('(','').replace(')',',')
#    for p in SP.lines:
#        coordIndex += '{} {} -1 '.format(p[0],p[1])
#    x3d_str = x3d_str.format(coordIndex, coordPoint, color[0], color[1], color[2])
#    return x3d_str

def make_x3d_schlegel_edges(schlegel_projection, scale_factor, radius, color):
    SP = schlegel_projection
    edges = SP.lines
    coords = list(map(lambda v: v*scale_factor,list(map(rnd_vec,SP.coordinates_of(SP.points)))))
    x3d_str = '\n'
    for line in SP.lines:
        edge = make_x3d_edge(coords[line[0]],coords[line[1]], radius, color)
        x3d_str += edge
    return x3d_str

def make_x3d_html(norm_ball, transparency=.7, online=False, directory=None, dual=False, show_labels=True, show_axes=True, highlight_edges=True, label_color=(1,1,1), edge_color=(1,1,1), label_size=1, edge_thickness=.01, vertex_size=.06, ball_color=(0,0,1), draw_schlegel_faces=False, window_size=(1000,700)):
    if dual:
        title = 'Dual TNorm Ball'
    else:
        title = 'TNorm Ball'
    text = ''
    P=norm_ball.polyhedron()

    width = window_size[0]
    height = window_size[1]

    if P.dim() == 4:

        facets = sorted(P.faces(3), key=lambda x: len(x.vertices()))

        # choose a face with maximal number of vertices. We will project from just outside this face.
        top_facet = facets[-1]
        tf_verts = [top_facet.vertices()[i].vector() for i in range(len(top_facet.vertices()))]
        vecs = [tf_verts[i]-tf_verts[0] for i in range(1, len(tf_verts))]
        N = normal(*span(vecs).basis())
        N_eps = .1*N

        # the center point of our projection face
        center = top_facet.as_polyhedron().center()

        # make sure N is an outward normal, and flip it if its not.
        in_, out_ = P.contains(center+N_eps), P.contains(center-N_eps)
        while not ( in_ or out_ ):
            N_eps *= .5
            in_, out_ = P.contains(center+N_eps), P.contains(center-N_eps)
        if in_:
            N = -N

        # we will rotate P so that N is in the direction of the positive w-axis, then project to the 
        # (x,y,z)-hyperplane. In retrospect this is probably not the most efficient approach. Should 
        # probably skip the rotation, and just project to the plane determined by N and passing through
        # the origin. 
        U = N.normalized()
        phi1, phi2, eta = get_hopf_coords(U,plane='xy')
        phi_xy = pi/2-phi1
        phi1, phi2, eta = get_hopf_coords(vector([0, n(sin(eta)), U[2], U[3]]), plane='yz')
        phi_yz = pi/2-phi1
        phi1, phi2, eta = get_hopf_coords(vector([0, 0, sin(eta), U[3]]), plane='xy')
        phi_zw = pi/2-phi2

        def rotated(vector4D):
            return rotate(vector4D, phi_xy, phi_yz, phi_zw)
        
        vertices = [v.vector() for v in P.vertices()]
        rotated_verts = list(map(rotated,vertices))
        rotated_center = rotated(center)

        # translate the polytope so that the center of the projection face is on the w-axis.
        translation = n(vector([rotated_center[0], rotated_center[1], rotated_center[2],0]))
        translated_verts = list(map(n,[v-translation for v in rotated_verts]))
        translated_center = rotated_center-translation

        # vertex indices of edges, faces, and facets (3d faces)
        edges = [(e.vertices()[0].index(), e.vertices()[1].index()) for e in P.faces(1)]
        faces = [tuple(f.vertices()[i].index() for i in range(len(f.vertices()))) for f in P.faces(2)] 
        facets = [tuple(f.vertices()[i].index() for i in range(len(f.vertices()))) for f in facets] 

        # we need to make sure the vertices of each face are cyclically ordered so x3dom can draw faces 
        fixed_faces = []
        for face in faces:
            face_verts = list(face)
            fixed_face = [face_verts.pop()]
            while len(face_verts) > 0:
                i=0
                while True:
                    if (fixed_face[-1],face_verts[i]) in edges or (face_verts[i],fixed_face[-1]) in edges:
                        fixed_face.append(face_verts[i])
                        face_verts.remove(face_verts[i])
                        break
                    else:
                        i += 1
            fixed_faces.append(fixed_face)
        faces = fixed_faces

        # We project from a point just above the top_facet. This point is chosen to be halfway between
        # the center point of the face (which is now on the w-axis), and the lowest point on the w-axis
        # that is above the center point, and that is in the hyperplane of one of the other facets.
        center_w_int = translated_center[3]
        w_intersections = []
        for f in facets[:-1]:
            vecs = [vertices[f[i]] - vertices[f[0]] for i in range(1,len(f))]
            N = rotated(normal(*span(vecs).basis()))
            k = N.dot_product(translated_verts[f[0]])
            if abs(N[3]) > 0.001:
                w_int = k/N[3]
                if w_int > center_w_int:
                    w_intersections.append(w_int)
        min_w_int = min(w_intersections)
        project_from_w = min((min_w_int + center_w_int)/2.0, center_w_int*2.0)
        project_from = vector([0,0,0,n(project_from_w)])

        # now we project all vertices to the (x,y,z)-hyperplane.
        projected_vertices = []
        for Q in translated_verts:
            PQ = Q - project_from
            t = -project_from[3]/PQ[3]
            projected_vert = project_from+t*PQ
            projected_vertices.append(n(projected_vert[:-1]))
        
        vertex_coords = projected_vertices

        ####### commented out becuase Sage's Schlegel projection does not work correctly as of (11/20/20)
        ####### it looks like this has been fixed and will be working in the next release, so we may want
        ####### to go back to using Sage's built in Schegel projection. So don't delte this...
        #SP = P.schlegel_projection()
        #hull=Polyhedron(SP.coordinates_of(SP.points))
        #coords = list(map(rnd_vec,SP.coordinates_of(SP.points)))
        #bounding_box = (vector([max([v[i] for v in coords]) for i in range(3)]),vector([min([v[i] for v in coords]) for i in range(3)]))

        # get the bounding box and from this a scale factor, which will allow us to put everything at a 
        # scale so default values for the size of edges, vertices, etc will always be reasonable.
        bounding_box = (vector([min([v[i] for v in vertex_coords]) for i in range(3)]),vector([max([v[i] for v in vertex_coords]) for i in range(3)]))
        scale_factor = n(3.0/(max(bounding_box[1])))

        # generate x3dom code for highlighted edges of norm ball
        edges_str = ''
        for e in edges:
            edge = make_x3d_edge(scale_factor*vertex_coords[e[0]],scale_factor*vertex_coords[e[1]], edge_thickness/2., edge_color)
            edges_str += edge
        #edges_str = make_x3d_schlegel_edges(SP, scale_factor, edge_thickness/2., edge_color)
        
        # generate x3dom code for 2D faces of the projected norm ball
        faces_str = ''
        if draw_schlegel_faces:
            coordIndex = ''
            scaled_coords = list(map(lambda v: v*scale_factor,list(map(rnd_vec,vertex_coords))))
            coordPoint = str(scaled_coords).replace('0.00000','0.0').replace('[(','').replace(')]','').replace(',','').replace('(','').replace(')',',')
            for face in faces:
                #vertices = [face.vertices()[i].vector() for i in range(len(face.vertices()))]
                #sp_indices = [SP.coords.index(v) for v in vertices]
                indices_str = ''
                for ind in face:
                    indices_str += '{} '.format(ind)
                indices_str += '-1, '
                coordIndex += indices_str
            coordIndex = coordIndex[:-2] #remove the extra comma at the end

            faces_str = '<Transform scale=\'1 1 1\'>\n<Shape>\n<IndexedFaceSet coordIndex=\'{0}\' solid=\'false\'>\n<Coordinate point=\'{1}\'>\n</Coordinate>\n</IndexedFaceSet>\n<Appearance>\n<Material ambientIntensity=\'.7\' transparency=\'{5}\' diffuseColor=\'{2} {3} {4}\' shininess=\'0.3\' specularColor=\'1.0 1.0 1.0\'>\n</Material>\n</Appearance>\n</Shape>\n</Transform>'.format(coordIndex, coordPoint, ball_color[0], ball_color[1], ball_color[2], transparency)


    else:
        bounding_box = P.bounding_box()
        max_coord = max(bounding_box[1])
        scale_factor = 3./max_coord
        faces_str = P.render_solid().transform(scale=(scale_factor,scale_factor,scale_factor)).x3d_str()
        face_str = fix_x3d_str(faces_str).replace('><','>\n<')
        faces_str = faces_str.replace('Material ','Material ambientIntensity=\'.7\' transparency=\'{}\' '.format(transparency))
        faces_str = faces_str.replace('specularColor=\'0.0 0.0 0.0\'','specularColor=\'1.0 1.0 1.0\'')
        faces_str = faces_str.replace('diffuseColor=\'0.0 0.0 1.0\'','diffuseColor=\'{} {} {}\''.format(ball_color[0], ball_color[1], ball_color[2]))
        faces_str = faces_str.replace('shininess=\'1.0\'','shininess=\'0.3\'')
        vertex_coords = P.vertices_list()

        edges_str = ''
        if highlight_edges:
            for e in P.faces(1):
                v1, v2 = e.vertices()[0].vector(), e.vertices()[1].vector()
                v1, v2 = n(v1,20)*scale_factor, n(v2,20)*scale_factor
                edges_str += make_x3d_edge(v1,v2,edge_thickness/2.,edge_color)

    vertices_str = ''
    labels_str = ''
    for i in range(len(vertex_coords)):
        v = vertex_coords[i]
        scaled_v = n(vector(v),20)*scale_factor
        normalized_v = n(vector(v),20).normalized()
        label_coords = scaled_v + normalized_v*.25
        v = norm_ball.vertex(i)
        if show_labels:
            if dual:
                label = '{}'.format(i)
                str_label_coords = '{} {} {}'.format(label_coords[0], label_coords[1], label_coords[2])
                labels_str += make_x3d_text(label,str_label_coords,'.3 .3 .3',label_size,label_color)
            else:
                labels_str += make_x3d_label(v.genus(),v.num_boundary_comps(),label_coords,'.3 .3 .3',label_size,.7*label_size,label_color)
        str_vert_coords = '{} {} {}'.format(scaled_v[0], scaled_v[1], scaled_v[2])
        vertices_str += make_x3d_vertex(str_vert_coords, VERTEX_RADIUS)


    xmin, xmax = n(bounding_box[0][0]*scale_factor,20), n(bounding_box[1][0]*scale_factor,20)
    ymin, ymax = n(bounding_box[0][1]*scale_factor,20), n(bounding_box[1][1]*scale_factor,20)
    zmin, zmax = n(bounding_box[0][2]*scale_factor,20), n(bounding_box[1][2]*scale_factor,20)

    if show_axes:
        datadir = os.path.dirname(__file__)
        path = os.path.join(datadir, 'data', 'x3d_templates', 'axes.html')
        with open(path,'r') as f:
            slack = min([(xmax-xmin)/2., (ymax-ymin)/2., (zmax-zmin)/2.])
            axis_len = ((xmax-xmin)+slack, (ymax-ymin)+slack, (zmax-zmin)+slack)
            axes = ''.join(f.readlines()).format(axis_len[1], axis_len[1]/2., axis_len[1]/2.+.2, axis_len[0], axis_len[0]/2., axis_len[0]/2.+.2,axis_len[2], axis_len[2]/2., axis_len[2]/2.+.2,AXIS_RADIUS, CONE_RADIUS, CONE_HEIGHT)
    else:
        axes = ''

    if online:
        head = "<head>\n<title>{}</title>\n<script type='text/javascript' src='http://www.x3dom.org/x3dom/release/x3dom.js'>\n</script>\n<link rel='stylesheet' type='text/css' href='http://www.x3dom.org/x3dom/release/x3dom.css'>\n</link>\n<h1>{}</h1>\n<style>\n</style>\n</head>\n".format(title,text)
    else:
        head = "<head>\n<title>{}</title>\n<script type='text/javascript' src=x3dom.js>\n</script>\n<link rel='stylesheet' type='text/css' href=x3dom.css></link>\n<h1>{}</h1>\n<style>\n</style>\n</head>\n".format(title,text) 




#    xaxes = "\n<Transform rotation='1 0 0 0' translation='0 0 0'>\n<Shape>\n<Appearance><Material ambientIntensity='0' emissiveColor='0 0 0' diffuseColor='0 0 0' specularColor='0 0 0' shininess='0.0078125' transparency='0'></Material></Appearance><IndexedLineSet coordIndex='0 1 -1 2 1 -1 3 1 -1' colorPerVertex='false'>\n<Coordinate point='{} 0 0, {} 0 0, '>\n</Coordinate>\n</IndexedLineSet>\n</Shape>\n</Transform>".format(xmin-2,xmax+2,)

    #light = "\n<directionalLight id='directional' direction='0 -1 0' on ='TRUE' intensity='2.0' shadowIntensity='0.0'>\n</directionalLight>"

    body = "<body>\n<x3d width='{}px' height='{}px', margin-left='300px'>\n<scene>\n<Background backUrl='[]' bind='false' bottomUrl='[]' crossOrigin='""' description='""' frontUrl='[]' groundAngle='[]' groundColor='(0.7,0.7,0.75)' isActive='false' leftUrl='[]' rightUrl='[]' scaling='[]' skyAngle='[]' skyColor='(0.7,0.7,0.75)' topUrl='[]' transparency='0/1' ></Background>\n{}\n</scene>\n</x3d>\n</body>\n".format(width,height,faces_str+labels_str+vertices_str+edges_str+axes)

    page = "<html>\n{}{}</html>".format(head,body)

    datadir = os.path.dirname(__file__)
    path = os.path.join(datadir, 'data', 'x3dom')
    css = os.path.join(path, 'x3dom.css')
    js = os.path.join(path, 'x3dom.js')

    shutil.copy2(css, directory)
    shutil.copy2(js, directory)

    filename = os.path.join(directory, 'normball_x3d.html')
    
    with open(filename,'w') as file:
        file.write(page)
    return os.path.abspath(filename)


def get_hopf_coords(vector, plane='xy'):
    x,y,z,w = n(vector[0]), n(vector[1]), n(vector[2]), n(vector[3])

    Pi = n(pi)

    if plane == 'xy':
        if abs(x) < 0.0001:
            phi = Pi/2 if y>0.000001 else -Pi/2
        else:
            phi = n(arctan(y/x)) + Pi*(x<0)
    
        if phi < 0:
            phi += 2.0*Pi
    
        if abs(z) < 0.0001:
            theta = Pi/2.0 if w>0.000001 else -Pi/2.0
        else:
            theta = n(arctan(w/z)) + Pi*(z<0)
    
        if theta < 0:
            theta += 2.0*Pi
    
        eta = n(arcsin(sqrt(x**2+y**2)))

    elif plane == 'yz':
        if abs(y) < 0.0001:
            phi = Pi/2.0 if z>0.000001 else -Pi/2.0
        else:
            phi = n(arctan(z/y)) + Pi*(y<0)
    
        if phi < 0:
            phi += 2.0*Pi
    
        if abs(x) < 0.0001:
            theta = Pi/2.0 if w>0.000001 else -Pi/2.0
        else:
            theta = n(arctan(w/x)) + Pi*(x<0)
    
        if theta < 0:
            theta += 2.0*Pi
    
        eta = n(arcsin(sqrt(y**2+z**2)))

    return n(phi), n(theta), n(eta)

# rotate vector4D by angle phi_ij about the ij-plane for ij in [xy, zw, xz, yw].
def rotate(vector4D, phi_xy, phi_yz, phi_zw):


    phi1, phi2, eta = get_hopf_coords(vector4D.normalized(), plane='xy')
    r = vector4D.norm()
    rotated_xy = vector([n(cos(phi1+phi_xy)*sin(eta)), n(sin(phi1+phi_xy)*sin(eta)), n(cos(phi2)*cos(eta)), n(sin(phi2)*cos(eta))])
    phi1, phi2, eta = get_hopf_coords(rotated_xy, plane='yz')
    rotated_yz = vector([n(cos(phi2)*cos(eta)), n(cos(phi1+phi_yz)*sin(eta)), n(sin(phi1+phi_yz)*sin(eta)), n(sin(phi2)*cos(eta))])
    phi1, phi2, eta = get_hopf_coords(rotated_yz, plane='xy')
    rotated_zw = vector([n(cos(phi1)*sin(eta)), n(sin(phi1)*sin(eta)), n(cos(phi2+phi_zw)*cos(eta)), n(sin(phi2+phi_zw)*cos(eta))])
    return rotated_zw*r

def translate(vector, translation_vector):
    pass

# get normal vector to the hyperplane containing four given points
#def normal(p1, p2, p3, p4):
#
#    N1 = Matrix([p2-p1, p3-p1, p4-p1, vector([1,0,0,0])]).determinant()
#    N2 = Matrix([p2-p1, p3-p1, p4-p1, vector([0,1,0,0])]).determinant()
#    N3 = Matrix([p2-p1, p3-p1, p4-p1, vector([0,0,1,0])]).determinant()
#    N4 = Matrix([p2-p1, p3-p1, p4-p1, vector([0,0,0,1])]).determinant()
#
#    return vector([N1,N2,N3,N4])
#
# get normal vector to the hyperplane containing four given points
def normal(v1, v2, v3):

    N1 = Matrix([v1, v2, v3, vector([1,0,0,0])]).determinant()
    N2 = Matrix([v1, v2, v3, vector([0,1,0,0])]).determinant()
    N3 = Matrix([v1, v2, v3, vector([0,0,1,0])]).determinant()
    N4 = Matrix([v1, v2, v3, vector([0,0,0,1])]).determinant()

    return vector([N1,N2,N3,N4])






#!/usr/bin/python

import threading
import platform
import os
import io
import sys

from tnorm.GUI.make_graphic import make_hasse, _show_polyhedron
from tnorm.GUI.make_table import make_qtons_table, make_all_vertices_table, make_facets_table, make_dnb_vertices_table
from tnorm.TN_wrapper import TN_wrapper
from tnorm.GUI.console import ConsoleText
from tnorm.utilities.hasse import get_hasse
from tnorm.GUI.canvas_image import CanvasImage
from tnorm.GUI.summary_page import generate_summary, generate_nb_overview, generate_dnb_overview
# background="..." doesn't work...

#import PIL
#from PIL import ImageTk
import snappy


FileType = io.TextIOWrapper if sys.version_info >= (3, 0) else file

try:
    import Tkinter as tk
    import tkFileDialog
    import tkMessageBox
except ImportError:  # Python 3.
    try:
        import tkinter as tk
        import tkinter.filedialog as tkFileDialog
        import tkinter.messagebox as tkMessageBox
    except ImportError:
        raise ImportError('Tkinter not available.')

try:
    import ttk as ttk
except ImportError:  # Python 3.
    try:
        from tkinter import ttk
    except ImportError:
        raise ImportError('Ttk not available.')

from tkinter import *

# Some constants.
if sys.platform in ['darwin']:
    COMMAND = {
        'new': 'Command+N',
        'open': 'Command+O',
        'save': 'Command+S',
        'close': 'Command+W',
        }
    COMMAND_KEY = {
        'new': '<Command-n>',
        'open': '<Command-o>',
        'save': '<Command-s>',
        'close': '<Command-w>',
        }
else:
    COMMAND = {
        'new': 'Ctrl+N',
        'open': 'Ctrl+O',
        'save': 'Ctrl+S',
        'close': 'Ctrl+W',
        }
    COMMAND_KEY = {
        'new': '<Control-n>',
        'open': '<Control-o>',
        'save': '<Control-s>',
        'close': '<Control-w>',
        }

class TNormApp:
    def __init__(self, parent):

         #------ constants for controlling layout of buttons ------
        BUTTON_WIDTH = 12
        BUTTON_YPAD = 0
        BG_COLOR = "#EAEAEA"


        self.parent = parent

        ### Our topmost frame is called MainFrame
        self.MainFrame = tk.Frame(parent) ###
        self.MainFrame.pack(expand=YES, fill=BOTH)

        #  main frame: control frame -----------------------------

        self.ControlFrame = tk.Frame(self.MainFrame, borderwidth=3,  relief=FLAT, bg=BG_COLOR, width=200) ###
        self.ControlFrame.pack(side=LEFT, expand=NO, fill=Y, anchor=W)

        # main frame: output frame ------------------------------

        self.OutputFrame = tk.Frame(self.MainFrame, bg=BG_COLOR,borderwidth=0) ###
        self.OutputFrame.pack(side=RIGHT, expand=YES, fill=BOTH)


        # control frame: load frame ------------------------------------------

        self.LoadFrame = tk.Frame(self.ControlFrame,bg=BG_COLOR,borderwidth=3)
        self.LoadFrame.pack(side=TOP, expand=NO, fill=X, padx=2, pady=2)

        # control frame: load frame: input box frame -------------------------------

        self.InputBoxFrame = tk.Frame(self.LoadFrame, bg=BG_COLOR)
        self.InputBoxFrame.pack(side=TOP)

        self.ManifoldLabel = tk.Label(self.InputBoxFrame, text='Manifold:', bg=BG_COLOR)
        self.ManifoldLabel.grid(row=0,column=0,sticky='w') 
        self.ManifoldEntry = tk.Entry(self.InputBoxFrame, width=10) 
        self.ManifoldEntry.bind("<Return>", self.load_on_enter)
        self.ManifoldEntry.grid(row=1, column=0)

        self.LoadButton = ttk.Button(self.InputBoxFrame, text="Load", command=self.load, padding=-8)
        self.LoadButton.grid(row=1,column=1)

        # control frame: load frame: load options frame ----------------------

        self.LoadOptionsFrame = tk.Frame(self.LoadFrame, bg=BG_COLOR)
        self.LoadOptionsFrame.pack(side=TOP,fill=X)

        self.allow_na_var = IntVar(None, 0)
        self.AllowsAdmissible = tk.Checkbutton(self.LoadOptionsFrame, bg=BG_COLOR, text='Allow non-admissible ', variable=self.allow_na_var)
        self.AllowsAdmissible.pack(side=TOP)

        self.force_simplicial = IntVar(None, 0)
        self.ForceSimplicial = tk.Checkbutton(self.LoadOptionsFrame, bg=BG_COLOR, text='Use simplicial H2 map', variable=self.force_simplicial)
        self.ForceSimplicial.pack(side=TOP)

        #self.BasisOption = tk.Frame(self.LoadOptionsFrame, bg=BG_COLOR)
        #self.BasisOption.pack(side=TOP, anchor=NW)
#
        #self.BasisLabel = tk.Label(self.BasisOption, bg=BG_COLOR, text='Peripheral curves:')
        #self.BasisLabel.grid(row=0,column=0,sticky='nw')
#
        #self.bdy_H1_basis_var = StringVar(None,'natural')
        #self.NaturalRadio = tk.Radiobutton(self.BasisOption, background=BG_COLOR, text='canonical (if link)', variable=self.bdy_H1_basis_var, value='natural')
        #self.NaturalRadio.grid(row=1, column=0, sticky='w', padx=10)
#
        #self.ShortestRadio = tk.Radiobutton(self.BasisOption, background=BG_COLOR, text='shortest', variable=self.bdy_H1_basis_var, value='shortest')
        #self.ShortestRadio.grid(row=2, column=0, sticky='w', padx=10)
        #self.bdy_H1_basis_var.set('natural')




        # control frame: buttons frame -------------------------------------
        self.ButtonsFrame = tk.Frame(self.ControlFrame, bg=BG_COLOR, borderwidth=0) ###
        self.ButtonsFrame.pack(side=TOP, expand=YES, fill=Y)


#        # control frame: buttons frame: view frame ------------------------------
#        self.ViewFrame = tk.Frame(self.ButtonsFrame, bg=BG_COLOR)
#        self.ViewFrame.pack(side=TOP, expand=NO, fill=X)
#
#        tk.Label(self.ViewFrame, text="View:", justify=LEFT, bg=BG_COLOR).pack(side=TOP, anchor=W)
#
#        self.SummaryButton = ttk.Button(self.ViewFrame, text="summary",command=self.show_summary,width=BUTTON_WIDTH)
#        self.SummaryButton.pack(pady = BUTTON_YPAD)
#
#        self.NormBallButton = ttk.Button(self.ViewFrame, text="norm ball",width=BUTTON_WIDTH)
#        self.NormBallButton.pack(pady = BUTTON_YPAD)
#
#        self.DualNormBallButton = ttk.Button(self.ViewFrame, text="dual norm ball",width=BUTTON_WIDTH)
#        self.DualNormBallButton.pack(pady = BUTTON_YPAD)
#        
#        self.QtonsInfo = ttk.Button(self.ViewFrame, text="qtons",width=BUTTON_WIDTH, command=self.show_all_qtons)
#        self.QtonsInfo.pack(pady = BUTTON_YPAD)
        # -----------------------------------------------------------------------


        #spacer ---------------------------------------------------------------------
        self.spacer1 = tk.Canvas(self.ButtonsFrame, height=10, bg=BG_COLOR, width=100, highlightthickness=-2)
        self.spacer1.pack()
        #----------------------------------------------------------------------------

        #control frame: buttons frame: options frame --------------------------------

        self.OptionsFrame = tk.Frame(self.ButtonsFrame, bg=BG_COLOR)
        self.OptionsFrame.pack(side=TOP, expand=YES, fill=BOTH)

        tk.Label(self.OptionsFrame, text="Options:", justify=LEFT, bg=BG_COLOR).pack(side=TOP, anchor=W)

        self.OverviewButton = ttk.Button(self.OptionsFrame, text="overview",command=self.show_overview,width=BUTTON_WIDTH)
        self.OverviewButton.pack(pady = BUTTON_YPAD)        

        self.VertexListButton = ttk.Button(self.OptionsFrame, text="vertex list",command=self.show_vertices,width=BUTTON_WIDTH)
        self.VertexListButton.pack(pady = BUTTON_YPAD)

        self.FaceListButton = ttk.Button(self.OptionsFrame, text="(n-1)-dim faces",width=BUTTON_WIDTH, command=self.show_faces)
        self.FaceListButton.pack(pady = BUTTON_YPAD)

        # control frame: buttons frame: options frame: facet button frame --------------
        self.FacetButtonFrame = tk.Frame(self.OptionsFrame, bg=BG_COLOR)
        self.FacetButtonFrame.pack(side=TOP, expand=NO)

        self.DimEntry = tk.Entry(self.FacetButtonFrame, width=2)
        self.DimEntry.pack(side=LEFT)

        self.FacetButton = ttk.Button(self.FacetButtonFrame, text='-dim facets', width=BUTTON_WIDTH-4, command=self.show_facets)
        self.FacetButton.pack(side=LEFT, pady = BUTTON_YPAD)
        # -------------------------------------------------------------------------------

#        self.spacer2 = tk.Canvas(self.ButtonsFrame, height=10, bg=BG_COLOR, width=100, highlightthickness=-2)
#        self.spacer2.pack()

        self.PolyhedronButton = ttk.Button(self.OptionsFrame, text="polyhedron",width=BUTTON_WIDTH, command=self.show_polyhedron)
        self.PolyhedronButton.pack(pady = BUTTON_YPAD)

        self.HasseButton = ttk.Button(self.OptionsFrame, text="hasse diagram",width=BUTTON_WIDTH, command=self.show_hasse)
        self.HasseButton.pack(pady = BUTTON_YPAD)




        # control frame: cancel button frame --------------

#        self.CancelButtonFrame = tk.Frame(self.ControlFrame, bg=BG_COLOR)
#        self.CancelButtonFrame.pack(side=BOTTOM, expand=NO, anchor=SW)

#        self.CancelButton = ttk.Button(self.CancelButtonFrame, text="Cancel", width=6, command=self.cancel_button_click)
#        self.CancelButton.pack(side=BOTTOM, pady=10)


        # main frame: output frame: notebook frame --------------------

        self.NotebookFrame = tk.Frame(self.OutputFrame, bg=BG_COLOR, pady=4)
        self.NotebookFrame.pack(side=TOP, expand=YES, fill=BOTH)  ###

        self.Notebook = Notebook = ttk.Notebook(self.NotebookFrame, padding=[0,0,7,0])

        self.SummaryTab = tk.Frame(self.NotebookFrame, bd=1, background='#DADADA')
        self.NormBallTab = tk.Frame(self.NotebookFrame, bd=1, background='#DADADA')
        self.DualNormBallTab = tk.Frame(self.NotebookFrame, bd=1, background='#DADADA')
        self.QtonsTab = tk.Frame(self.NotebookFrame, bd=1, background='#DADADA')
        self.HelpTab = tk.Frame(self.NotebookFrame, bd=1, background='#DADADA')
        Notebook.add(self.SummaryTab, text='Summary', sticky='nsew', padding=[0,0,0,0], state='disabled')
        Notebook.add(self.NormBallTab, text='Norm ball', sticky='nsew', padding=[0,0,0,0], state='disabled')
        Notebook.add(self.DualNormBallTab, text='Dual norm ball', sticky='nsew', padding=[0,0,0,0], state='disabled')
        Notebook.add(self.QtonsTab, text='qton surfaces', sticky='nsew', padding=[0,0,0,0], state='disabled')
        Notebook.add(self.HelpTab, text='Help', sticky='nsew', padding=[0,0,0,0], state='normal')
        Notebook.grid(row=0, column=0, sticky=tk.NSEW, padx=0, pady=0)
        self.NotebookFrame.grid_columnconfigure(0, weight=1)
        self.NotebookFrame.grid_rowconfigure(0, weight=1)
        self.disable_all_buttons()
        Notebook.bind('<<NotebookTabChanged>>', self.update_current_tab)
        self.tabs = [self.SummaryTab, self.NormBallTab, self.DualNormBallTab, self.QtonsTab, self.HelpTab]

        # main frame: output frame: status frame -----------------

        self.StatusFrame = tk.Frame(self.OutputFrame,borderwidth=0,
            height=120, bg=BG_COLOR, pady=6) ###
        self.StatusFrame.pack(side=TOP, fill=X)

        # main frame: output frame: buttons frame: icon frame -------------------------------

        self.IconFrame = tk.Frame(self.ButtonsFrame, bg=BG_COLOR)
        self.IconFrame.pack(side=TOP, expand=YES)

        datadir = os.path.dirname(__file__)
        path = os.path.join(datadir, 'images', 'spinpoly', 'poly{}.gif')
        self.icon_frames = [PhotoImage(file=path.format(i)) for i in range(9)]

        self.IconGif = Label(self.IconFrame,image=self.icon_frames[0], bd=-2)
        self.IconGif.pack(side=TOP)

        # main frame: output frame: status frame: console frame ---------------------------------------------------

        self.ConsoleFrame = tk.Frame(self.StatusFrame, borderwidth=2,
            height=120, relief=RIDGE) ###
        self.ConsoleFrame.grid(row=0,column=0, sticky='nsew')

        self.ConsolePadding = tk.Frame(self.StatusFrame, width=8, bg=BG_COLOR)
        self.ConsolePadding.grid(row=0,column=1)
        self.StatusFrame.grid_columnconfigure(0, weight=1)

        self.console = ConsoleText(self.ConsoleFrame, height=4, width=20)
        self.console.pack(side=LEFT, expand=YES,fill=X)
        self.console.start()
        self.ball = None
        self.dual_ball = None

        s = ttk.Style()
        s.configure("Treeview.Heading", relief='flat')

        if platform.system() == 'Darwin':
            s.theme_settings("default", {"TNotebook.Tab": {"configure": {"padding": [30, 30]}}})
        s.map("TButton",
            fieldbackground=[('disabled', '#F0F0F0')], foreground=[('disabled', '#909090')])
        s.map("TNotebook.Tab",
            fieldbackground=[('disabled', '#F0F0F0')], foreground=[('disabled', '#909090')])

        self.show_welcome()

    def current_tab(self):
        if self.Notebook.select() != '':
            tab_name = self.Notebook.tab(self.Notebook.select(),'text')
            if tab_name == 'Summary':
                return self.SummaryTab
            elif tab_name == 'Norm ball':
                return self.NormBallTab
            elif tab_name == 'Dual norm ball':
                return self.DualNormBallTab
            elif tab_name == 'qton surfaces':
                return self.QtonsTab
            elif tab_name == 'Help':
                return self.HelpTab
        else:
            return None

    def update_current_tab(self, event=None):
        self.parent.update_idletasks()
        tab = self.current_tab()
        if tab == self.SummaryTab:
            self.disable([self.DimEntry, self.OverviewButton, self.VertexListButton, self.FaceListButton, self.FacetButton, self.PolyhedronButton, self.HasseButton])            
        if tab == self.NormBallTab:
            self.DimEntry.delete(0,END)
            self.enable([self.DimEntry, self.OverviewButton, self.VertexListButton, self.FaceListButton, self.FacetButton, self.PolyhedronButton, self.HasseButton])
            if len(self.NormBallTab.winfo_children()) != 0:
                pass
            else:
                self.show_overview()
        elif tab == self.DualNormBallTab:
            self.enable([self.DimEntry, self.OverviewButton, self.VertexListButton, self.FaceListButton, self.FacetButton, self.PolyhedronButton, self.HasseButton])
            self.disable([self.DimEntry, self.FaceListButton, self.FacetButton])
            self.DimEntry.delete(0,END)
            if len(self.DualNormBallTab.winfo_children()) != 0:
                pass
            else:
                self.show_overview()

        elif tab == self.QtonsTab:
            self.disable([self.DimEntry, self.OverviewButton, self.VertexListButton, self.FaceListButton, self.FacetButton, self.PolyhedronButton, self.HasseButton])
            if len(self.QtonsTab.winfo_children()) == 0:
                self.show_all_qtons()
            else:
                pass
        elif tab == self.HelpTab:
            self.disable([self.DimEntry, self.VertexListButton, self.FaceListButton, self.FacetButton, self.PolyhedronButton, self.HasseButton])

        self.parent.update_idletasks()

    # disable tabs and option buttons

    def disable_all_tabs(self):
        current_tab = self.current_tab()
        # need to disable the current tab last, to avoid invoking update_current_tab().
        for tab in self.tabs:
            if tab != current_tab:
                self.Notebook.tab(tab, state='disabled')
        if current_tab != None:
            self.Notebook.tab(current_tab, state='disabled')


    def disable_all_buttons(self):
        self.VertexListButton.configure(state='disabled')
        self.FaceListButton.configure(state='disabled')
        self.FacetButton.configure(state='disabled')
        self.PolyhedronButton.configure(state='disabled')
        self.HasseButton.configure(state='disabled')
        self.DimEntry.configure(state='disabled')        

    def tabs_normal(self):
        self.Notebook.tab(self.SummaryTab, state='normal')
        self.Notebook.tab(self.NormBallTab, state='normal')
        self.Notebook.tab(self.DualNormBallTab, state='normal')
        self.Notebook.tab(self.QtonsTab, state='normal')
        self.Notebook.tab(self.HelpTab, state='normal')

    def enable(self, widgets):
        for widget in widgets:
            widget.configure(state='normal')
    
    def disable(self, widgets):
        for widget in widgets:
            widget.configure(state='disabled')


    # clear tab to load new table/graphic/etc ------------------------

    def clear_all_tabs(self):
        for tab in [self.SummaryTab, self.NormBallTab, self.DualNormBallTab, self.QtonsTab]:
            self.clear_tab(tab)

    def clear_tab(self, tab):
        for widget in tab.winfo_children():
            widget.destroy()


    # spinning polyhedron -----------------------------------------------

    def start_spin(self):
        self.computing = True
        self.IconFrame.after(0, self.update, 0)
        
    def update(self,ind):
        if self.computing == True:
            frame = self.icon_frames[ind]
            ind = (ind+1)%9
            self.IconGif.configure(image=frame)
            self.IconFrame.after(100, self.update, ind)
        else:
            return None

    def stop_spin(self):
        self.computing = False


    # load button commands -------------------------------------------

    def load(self):
        try:
            M_str = self.ManifoldEntry.get()
            computing=True
            M = snappy.Manifold(M_str)
            t = threading.Thread(target=self.load_computations, args=(M,))
            t.start()
            self.start_spin()

        except IOError:
            print("invalid entry")

    def load_computations(self, M):
#        try:
        #self.disable([self.LoadButton, self.ManifoldEntry, self.NaturalRadio, self.ShortestRadio, self.AllowsAdmissible, self.BasisLabel, self.ForceSimplicial, self.OverviewButton])
        self.disable([self.LoadButton, self.ManifoldEntry, self.AllowsAdmissible, self.ForceSimplicial, self.OverviewButton])
        self.disable_all_buttons()
        self.disable_all_tabs()
        self.clear_all_tabs()
        self.wrapper = TN_wrapper(M, allows_non_admissible=bool(self.allow_na_var.get()), force_simplicial_homology=bool(self.force_simplicial.get()))
        self.ball = self.wrapper.norm_ball
        self.dual_ball = self.wrapper.dual_norm_ball
        generate_summary(self)
        self.tabs_normal()
        #self.enable([self.LoadButton, self.ManifoldEntry, self.NaturalRadio, self.ShortestRadio, self.AllowsAdmissible, self.ForceSimplicial, self.BasisLabel])
        self.enable([self.LoadButton, self.ManifoldEntry, self.AllowsAdmissible, self.ForceSimplicial])
        self.Notebook.select(self.SummaryTab)
#        except Exception as e: print('Error: {}'.format(e))
        self.stop_spin()

    # option button commands ----------------------------------------------------

    def show_overview(self):
        tab = self.current_tab()
        if tab == self.HelpTab:
            self.show_welcome()
        elif tab == self.NormBallTab:
            self.show_nb_overview()
        elif tab == self.DualNormBallTab:
            self.show_dnb_overview()

    def show_welcome(self):
        datadir = os.path.dirname(__file__)
        path = os.path.join(datadir,'data','welcome.txt')
        with open(path, 'r') as file:
            self.welcome = tk.Text(self.HelpTab, wrap='none', bg="#DADADA", fg="#444444", height=34, width=103, borderwidth=-2, padx=-1, pady=-1, insertwidth=0)
            self.welcome.grid(row=0, column=0, sticky='nswe')
            for line in file.readlines():
                self.welcome.insert(tk.END, line)

    def show_nb_overview(self):
        generate_nb_overview(self)

    def show_dnb_overview(self):
        generate_dnb_overview(self)



    def show_summary(self):
        computing=True
        t = threading.Thread(target=generate_summary,args=(self,))
        t.start()
        self.start_spin()

    def show_vertices(self):
        tab = self.current_tab()
        self.clear_tab(tab)
        computing = True
        if tab == self.NormBallTab:
            t = threading.Thread(target=make_all_vertices_table, args=(self,tab))
        else:
            t = threading.Thread(target=make_dnb_vertices_table, args=(self,tab))
        t.start()
        self.start_spin()


    def show_faces(self):
        P = self.ball.polyhedron()
        dim = P.dim()-1
        make_facets_table(self, self.NormBallTab, dim)


    def show_facets(self):
        computing = True
        self.clear_tab(self.NormBallTab)
        dim = int(self.DimEntry.get())
        t = threading.Thread(target=make_facets_table, args=(self, self.NormBallTab, dim))
        t.start()
        self.start_spin()


    def show_all_qtons(self):
        computing = True
        self.clear_tab(self.QtonsTab)
        t = threading.Thread(target=make_qtons_table,args=(self,))
        t.start()
        self.start_spin()

    def show_hasse(self):
        computing = True
        tab = self.current_tab()
        self.clear_tab(tab)
        t = threading.Thread(target=make_hasse, args=(self, tab))
        t.start()
        self.start_spin()

    def show_schlegel(self):
        computing = True
        self.clear_canvas()
        t = threading.Thread(target=make_schlegel, args=(self,))
        t.start()
        self.start_spin()

    def show_polyhedron(self):
        computing = True
        online = False
        tab = self.current_tab()
        self.clear_tab(tab)
        t = threading.Thread(target=_show_polyhedron, args=(self, online, tab))
        t.start()
        self.start_spin()


#    def cancel_button_click(self):
#        self.parent.destroy()


    # button bindings for Enter key ----------------------------------

    def load_on_enter(self,event):
        self.load()


def start():
    root = tk.Tk()
    root.minsize(500,500)
    root.geometry("900x600")
    root.title('tnorm')
    myapp = TNormApp(root)
    datadir = os.path.dirname(__file__)
    icon_path = os.path.join(datadir, 'images', 'icon', 'icon.png')
    img = tk.PhotoImage(file=icon_path)
    root.tk.call('wm', 'iconphoto', root._w, img)
    root.mainloop()



if __name__ == '__main__':
    start()





from .sage_types import Integer, QQ, Matrix



### return True if vertex "vert" of tetrahedron "tet" is in given cusp
def in_cusp(T,tet,vert,cusp):
    return T.tetrahedron(int(tet)).vertex(int(vert)) == T.cusp(int(cusp)).vertex()



### convert regina types (integers, rational, matrices) to sage types
def regina_to_sage_int(q):
    return Integer(q.longValue())

def regina_to_sage_rat(q):
    return QQ((q.numerator().longValue(),q.denominator().longValue()))

def regina_to_sage_mat(q):
    return Matrix([[regina_to_sage_int(q.entry(i,j)) for j in range(q.columns())] for i in range(q.rows())])



def get_oriented_quads(oriented_spun_surface,tet,v1,v2):
    assert v1 < v2
    s = oriented_spun_surface
    tet = int(tet)
    v2 = int(v2)
    if v1 == 0:
        return regina_to_sage_int(s.orientedQuads(tet,v2-int(1),True))
    elif v1 == 2:
        return regina_to_sage_int(s.orientedQuads(tet,int(0),False))
    elif v2 == 2:
        return regina_to_sage_int(s.orientedQuads(tet,int(2),False))
    else:
        return regina_to_sage_int(s.orientedQuads(tet,int(1),False))
    return "invalid input"

def get_quads(spun_surface,tet,v1,v2):
    assert v1 < v2
    assert v1 == 0
    s = spun_surface
    tet = int(tet)
    v2 = int(v2)
    return regina_to_sage_int(s.quads(tet,v2-int(1)))





from tnorm.version import version, __version__

from tnorm.TN_wrapper import TN_wrapper as load

import tnorm.kernel
import tnorm.extras
from tnorm.test import test
#import tnorm.GUI


#IN_SAGE=False




from tnorm.utilities.sage_types import Integer, Matrix, vector

from tnorm.utilities.regina_helpers import regina_to_sage_int, regina_to_sage_mat


### compute the euler characteristic. If no angle structure is given, computes
### an angle structure to use for computation of Euler characteristic (i.e., the 
### second argument is optional). 
def euler_char_(spun_surface,angle_struct_matrix=None):
    T=spun_surface.triangulation()
    if angle_struct_matrix==None:
        angles = solve_lin_gluing_eq(T)
    else:
        angles = angle_struct_matrix
    tets = T.size()
    ec = 0
    for i in range(tets):
        for j in range(3):
            ec -= angles[i][j]*regina_to_sage_int(spun_surface.quads(int(i),int(j)))
    return ec


##############-----------------------------------------------------------

### solve the linear gluing equations, including the linear part of the holonomy
### equations for meridians and longitudes. Returns a sage Matrix.
def solve_lin_gluing_eq(T):
    GE = regina_to_sage_mat(T.gluingEquations())
    for i in range(T.size()):
        # append equations log(z1)+log(z2)+log(z3)=\pi
        GE = GE.insert_row(GE.dimensions()[0],[Integer(0) for j in range(3*i)]+[Integer(1) for j in range(3)] + [Integer(0) for j in range(3*T.size()-3*(i+1))])
        
        ##### for sympy implementation
        #GE = GE.col_join((Matrix([[Integer(0) for j in range(3*i)]+[Integer(1) for j in range(3)] + [Integer(0) for j in range(3*T.size()-3*(i+1))]])))
    edges = T.countEdges()
    cusps = T.countCusps()
    # 2pi for edge equations, 0 for cusp equations, pi for three dihedral angles of a tetrahedron (then divide out pi)
    v= vector([2 for i in range(edges)]+[0 for i in range(2*cusps)]+[1 for i in range(T.size())])

    assert len(v) == len(GE.rows())
    angles_vec = GE.solve_right(v)

###### for sympy implementation    
#    assert v.rows == GE.rows
#    solutions, params = GE.gauss_jordan_solve(v)  # solve the system of equations, returns parametrization of solutions
#    param_zeroes = { tau:0 for tau in params } #set parameters to 0 to get a solution which will hopefully be an integer solution.
#    angles_vec = solutions.xreplace(param_zeroes) 
    
    angles = Matrix([[angles_vec[j+i] for i in range(3)] for j in range(0,len(angles_vec),3)])
    return angles











### It appears none of the below functions are being used.
##############-----------------------------------------------------------

def angle_struct_to_vec(regina_structure):
    tets = regina_structure.triangulation().size()
    return [regina.engine.Rational(regina_structure.angle(i,j)) for i in range(tets) for j in range(3)]

def eval_angle_struct(T,angle_struct):
    ge = T.gluingEquations()
    asv = angle_struct_to_vec(angle_struct)
    tets = T.size()
    edges = T.countEdges()
    cusps = T.countCusps()
    result = [regina.engine.Rational(0) for i in range(edges+2*cusps)]
    for i in range(edges+2*cusps):
        for j in range(3*tets):
            result[i] += regina.engine.Rational(ge.entry(i,j))*asv[j]
    return result

def find_good_structures(T,structures):
    edges = T.countEdges()
    cusps = T.countCusps()
    good_structures = []
    good = [regina.engine.Rational(2) for i in range(edges)]+[regina.engine.Rational(0) for i in range(2*cusps)]
    for i in range(structures.size()):
        if eval_angle_struct(T,structures.structure(i)) == good:
            good_structures.append(structures.structure(i))
    return good_structures

#############-------------------------------------------------------------------



# -*- coding: utf-8 -*-
# Advanced zoom for images of various types from small to huge up to several GB
import math
import warnings

import tkinter as tk
from tkinter import ttk
from PIL import Image, ImageTk

from tnorm.GUI.scroll_bars import AutoScrollbar

class CanvasImage:
    """ Display and zoom image """
    def __init__(self, tnorm_app, path):
        """ Initialize the ImageFrame """
        self.imscale = 1.0  # scale for the canvas image zoom, public for outer classes
        self.__delta = 1.2  # zoom magnitude
        self.__filter = Image.ANTIALIAS  # could be: NEAREST, BILINEAR, BICUBIC and ANTIALIAS
        self.__previous_state = 0  # previous state of the keyboard
        self.path = path  # path to the image, should be public for outer classes
        # Create ImageFrame in placeholder widget

        self.__imframe = ttk.Frame(tnorm_app.ImageFrame)  # placeholder of the ImageFrame object
        # Vertical and horizontal scrollbars for canvas
        hbar = AutoScrollbar(self.__imframe, orient='horizontal')
        vbar = AutoScrollbar(self.__imframe, orient='vertical')
        hbar.grid(row=1, column=0, sticky='we')
        vbar.grid(row=0, column=1, sticky='ns')

        #zoom buttons
#        self.zoom_frame = ttk.Frame(self.__imframe)
#        self.zoom_frame.grid(row=0, column=0, rowspan=2, sticky='ns')
        self.z_plus = ttk.Button(tnorm_app.ZoomFrame, text='+', width=1, command=self.__zoom_in, padding=[-20,0,-20,0])
#        self.z_plus.bind("<Button-1>",self.__zoom_in)
        self.z_plus.grid(row=1, column=0, sticky='n')
        self.z_minus = ttk.Button(tnorm_app.ZoomFrame, text='-', width=1, command=self.__zoom_out, padding=[-20,0,-20,0])
#        self.z_plus.bind("<Button-1>",self.__zoom_out)
        self.z_minus.grid(row=2, column=0, sticky='s')
        tnorm_app.ZoomFrame.rowconfigure(0, weight=1)
        tnorm_app.ZoomFrame.rowconfigure(3, weight=1)



        # Create canvas and bind it with scrollbars. Public for outer classes
        self.canvas = tk.Canvas(self.__imframe, highlightthickness=0,
                                xscrollcommand=hbar.set, yscrollcommand=vbar.set)
        self.canvas.grid(row=0, column=0, sticky='nswe')
        self.canvas.update()  # wait till canvas is created
        hbar.configure(command=self.__scroll_x)  # bind scrollbars to the canvas
        vbar.configure(command=self.__scroll_y)
        # Bind events to the Canvas
        self.canvas.bind('<Configure>', lambda event: self.__show_image())  # canvas is resized
        self.canvas.bind('<ButtonPress-1>', self.__move_from)  # remember canvas position
        self.canvas.bind('<B1-Motion>',     self.__move_to)  # move canvas to the new position
        self.canvas.bind('<MouseWheel>', self.__wheel)  # zoom for Windows and MacOS, but not Linux
        self.canvas.bind('<Button-5>',   self.__wheel)  # zoom for Linux, wheel scroll down
        self.canvas.bind('<Button-4>',   self.__wheel)  # zoom for Linux, wheel scroll up
#        self.canvas.bind('<Key-43>',self.__keystroke)
#        self.canvas.bind('<Key-45>',self.__keystroke)
        # Handle keystrokes in idle mode, because program slows down on a weak computers,
        # when too many key stroke events in the same time
        self.canvas.bind('<Key>', lambda event: self.canvas.after_idle(self.__keystroke, event))
#        self.canvas.bind('<Key>', self.__keystroke)
        # Decide if this image huge or not
        self.__huge = False  # huge or not
        self.__huge_size = 14000  # define size of the huge image
        self.__band_width = 1024  # width of the tile band
        Image.MAX_IMAGE_PIXELS = 1000000000  # suppress DecompressionBombError for the big image
        with warnings.catch_warnings():  # suppress DecompressionBombWarning
            warnings.simplefilter('ignore')
            self.__image = Image.open(self.path)  # open image, but down't load it
        self.imwidth, self.imheight = self.__image.size  # public for outer classes
        if self.imwidth * self.imheight > self.__huge_size * self.__huge_size and \
           self.__image.tile[0][0] == 'raw':  # only raw images could be tiled
            self.__huge = True  # image is huge
            self.__offset = self.__image.tile[0][2]  # initial tile offset
            self.__tile = [self.__image.tile[0][0],  # it have to be 'raw'
                           [0, 0, self.imwidth, 0],  # tile extent (a rectangle)
                           self.__offset,
                           self.__image.tile[0][3]]  # list of arguments to the decoder
        self.__min_side = min(self.imwidth, self.imheight)  # get the smaller image side
        # Create image pyramid
        self.__pyramid = [self.smaller()] if self.__huge else [Image.open(self.path)]
        # Set ratio coefficient for image pyramid
        self.__ratio = max(self.imwidth, self.imheight) / self.__huge_size if self.__huge else 1.0
        self.__curr_img = 0  # current image from the pyramid
        self.__scale = self.imscale * self.__ratio  # image pyramide scale
        self.__reduction = 2  # reduction degree of image pyramid
        w, h = self.__pyramid[-1].size
        while w > 512 and h > 512:  # top pyramid image is around 512 pixels in size
            w /= self.__reduction  # divide on reduction degree
            h /= self.__reduction  # divide on reduction degree
            self.__pyramid.append(self.__pyramid[-1].resize((int(w), int(h)), self.__filter))
        # Put image into container rectangle and use it to set proper coordinates to the image
        self.container = self.canvas.create_rectangle((0, 0, self.imwidth, self.imheight), width=0)
        
        self.__show_image()
        self.canvas.focus_set()  # set focus on the canvas

        # Below is a hack to get image to load at (approximately) canvas size.
        tnorm_app.parent.update_idletasks()
        window_width = float(tnorm_app.parent.winfo_width())
        canvas_width = window_width-227
        total_scale = canvas_width/self.imwidth
        scale=1
        while scale > total_scale:
            self.__zoom_out()
            scale *= 1/self.__delta

    def smaller(self):
        """ Resize image proportionally and return smaller image """
        w1, h1 = float(self.imwidth), float(self.imheight)
        w2, h2 = float(self.__huge_size), float(self.__huge_size)
        aspect_ratio1 = w1 / h1
        aspect_ratio2 = w2 / h2  # it equals to 1.0
        if aspect_ratio1 == aspect_ratio2:
            image = Image.new('RGB', (int(w2), int(h2)))
            k = h2 / h1  # compression ratio
            w = int(w2)  # band length
        elif aspect_ratio1 > aspect_ratio2:
            image = Image.new('RGB', (int(w2), int(w2 / aspect_ratio1)))
            k = h2 / w1  # compression ratio
            w = int(w2)  # band length
        else:  # aspect_ratio1 < aspect_ration2
            image = Image.new('RGB', (int(h2 * aspect_ratio1), int(h2)))
            k = h2 / h1  # compression ratio
            w = int(h2 * aspect_ratio1)  # band length
        i, j, n = 0, 1, round(0.5 + self.imheight / self.__band_width)
        while i < self.imheight:
            print('\nOpening image: {j} from {n}'.format(j=j, n=n))
            band = min(self.__band_width, self.imheight - i)  # width of the tile band
            self.__tile[1][3] = band  # set band width
            self.__tile[2] = self.__offset + self.imwidth * i * 3  # tile offset (3 bytes per pixel)
            self.__image.close()
            self.__image = Image.open(self.path)  # reopen / reset image
            self.__image.size = (self.imwidth, band)  # set size of the tile band
            self.__image.tile = [self.__tile]  # set tile
            cropped = self.__image.crop((0, 0, self.imwidth, band))  # crop tile band
            image.paste(cropped.resize((w, int(band * k)+1), self.__filter), (0, int(i * k)))
            i += band
            j += 1
        print('\r' + 30*' ' + '\r')  # hide printed string
        return image

    def redraw_figures(self):
        """ Dummy function to redraw figures in the children classes """
        pass

    def grid(self, **kw):
        """ Put CanvasImage widget on the parent widget """
        self.__imframe.grid(**kw)  # place CanvasImage widget on the grid
        self.__imframe.grid(sticky='nswe')  # make frame container sticky
        self.__imframe.rowconfigure(0, weight=1)  # make canvas expandable
        self.__imframe.columnconfigure(0, weight=1)

    def pack(self, **kw):
        """ Exception: cannot use pack with this widget """
        raise Exception('Cannot use pack with the widget ' + self.__class__.__name__)

    def place(self, **kw):
        """ Exception: cannot use place with this widget """
        raise Exception('Cannot use place with the widget ' + self.__class__.__name__)

    # noinspection PyUnusedLocal
    def __scroll_x(self, *args, **kwargs):
        """ Scroll canvas horizontally and redraw the image """
        self.canvas.xview(*args)  # scroll horizontally
        self.__show_image()  # redraw the image

    # noinspection PyUnusedLocal
    def __scroll_y(self, *args, **kwargs):
        """ Scroll canvas vertically and redraw the image """
        self.canvas.yview(*args)  # scroll vertically
        self.__show_image()  # redraw the image

    def __show_image(self):
        """ Show image on the Canvas. Implements correct image zoom almost like in Google Maps """
        box_image = self.canvas.coords(self.container)  # get image area
        box_canvas = (self.canvas.canvasx(0),  # get visible area of the canvas
                      self.canvas.canvasy(0),
                      self.canvas.canvasx(self.canvas.winfo_width()),
                      self.canvas.canvasy(self.canvas.winfo_height()))
        box_img_int = tuple(map(int, box_image))  # convert to integer or it will not work properly
        # Get scroll region box
        box_scroll = [min(box_img_int[0], box_canvas[0]), min(box_img_int[1], box_canvas[1]),
                      max(box_img_int[2], box_canvas[2]), max(box_img_int[3], box_canvas[3])]
        # Horizontal part of the image is in the visible area
        if  box_scroll[0] == box_canvas[0] and box_scroll[2] == box_canvas[2]:
            box_scroll[0]  = box_img_int[0]
            box_scroll[2]  = box_img_int[2]
        # Vertical part of the image is in the visible area
        if  box_scroll[1] == box_canvas[1] and box_scroll[3] == box_canvas[3]:
            box_scroll[1]  = box_img_int[1]
            box_scroll[3]  = box_img_int[3]
        # Convert scroll region to tuple and to integer
        self.canvas.configure(scrollregion=tuple(map(int, box_scroll)))  # set scroll region
        x1 = max(box_canvas[0] - box_image[0], 0)  # get coordinates (x1,y1,x2,y2) of the image tile
        y1 = max(box_canvas[1] - box_image[1], 0)
        x2 = min(box_canvas[2], box_image[2]) - box_image[0]
        y2 = min(box_canvas[3], box_image[3]) - box_image[1]
        if int(x2 - x1) > 0 and int(y2 - y1) > 0:  # show image if it in the visible area
            if self.__huge and self.__curr_img < 0:  # show huge image
                h = int((y2 - y1) / self.imscale)  # height of the tile band
                self.__tile[1][3] = h  # set the tile band height
                self.__tile[2] = self.__offset + self.imwidth * int(y1 / self.imscale) * 3
                self.__image.close()
                self.__image = Image.open(self.path)  # reopen / reset image
                self.__image.size = (self.imwidth, h)  # set size of the tile band
                self.__image.tile = [self.__tile]
                image = self.__image.crop((int(x1 / self.imscale), 0, int(x2 / self.imscale), h))
            else:  # show normal image
                image = self.__pyramid[max(0, self.__curr_img)].crop(  # crop current img from pyramid
                                    (int(x1 / self.__scale), int(y1 / self.__scale),
                                     int(x2 / self.__scale), int(y2 / self.__scale)))
            #
            imagetk = ImageTk.PhotoImage(image.resize((int(x2 - x1), int(y2 - y1)), self.__filter))
            imageid = self.canvas.create_image(max(box_canvas[0], box_img_int[0]),
                                               max(box_canvas[1], box_img_int[1]),
                                               anchor='nw', image=imagetk)
            self.canvas.lower(imageid)  # set image into background
            self.canvas.imagetk = imagetk  # keep an extra reference to prevent garbage-collection

    def __move_from(self, event):
        """ Remember previous coordinates for scrolling with the mouse """
        self.canvas.scan_mark(event.x, event.y)

    def __move_to(self, event):
        """ Drag (move) canvas to the new position """
        self.canvas.scan_dragto(event.x, event.y, gain=1)
        self.__show_image()  # zoom tile and show it on the canvas

    def outside(self, x, y):
        """ Checks if the point (x,y) is outside the image area """
        bbox = self.canvas.coords(self.container)  # get image area
        if bbox[0] < x < bbox[2] and bbox[1] < y < bbox[3]:
            return False  # point (x,y) is inside the image area
        else:
            return True  # point (x,y) is outside the image area

    def __wheel(self, event):
        """ Zoom with mouse wheel """
        x = self.canvas.canvasx(event.x)  # get coordinates of the event on the canvas
        y = self.canvas.canvasy(event.y)
        if self.outside(x, y): return  # zoom only inside image area
        scale = 1.0
        # Respond to Linux (event.num) or Windows (event.delta) wheel event
        if event.num == 5 or event.delta == -120 or event.keycode == 45:  # scroll down, smaller
            if round(self.__min_side * self.imscale) < 30: return  # image is less than 30 pixels
            self.imscale /= self.__delta
            scale        /= self.__delta
        if event.num == 4 or event.delta == 120 or event.keycode == 43:  # scroll up, bigger
            i = min(self.canvas.winfo_width(), self.canvas.winfo_height()) >> 1
            if i < self.imscale: return  # 1 pixel is bigger than the visible area
            self.imscale *= self.__delta
            scale        *= self.__delta
        # Take appropriate image from the pyramid
        k = self.imscale * self.__ratio  # temporary coefficient
        self.__curr_img = min((-1) * int(math.log(k, self.__reduction)), len(self.__pyramid) - 1)
        self.__scale = k * math.pow(self.__reduction, max(0, self.__curr_img))
        #
        self.canvas.scale('all', x, y, scale, scale)  # rescale all objects
        # Redraw some figures before showing image on the screen
        self.redraw_figures()  # method for child classes
        self.__show_image()

    def __zoom_in(self):
        """ Zoom in with + key """
#        x = self.canvas.canvasx(event.x)  # get coordinates of the event on the canvas
#        y = self.canvas.canvasy(event.y)
#        if self.outside(x, y): return  # zoom only inside image area
        scale = 1.0
        i = min(self.canvas.winfo_width(), self.canvas.winfo_height()) >> 1
        if i < self.imscale: return  # 1 pixel is bigger than the visible area
        self.imscale *= self.__delta
        scale        *= self.__delta
        # Take appropriate image from the pyramid
        k = self.imscale * self.__ratio  # temporary coefficient
        self.__curr_img = min((-1) * int(math.log(k, self.__reduction)), len(self.__pyramid) - 1)
        self.__scale = k * math.pow(self.__reduction, max(0, self.__curr_img))
        #
        self.canvas.scale('all', .5, .5, scale, scale)  # rescale all objects
        # Redraw some figures before showing image on the screen
        self.redraw_figures()  # method for child classes
        self.__show_image()

    def __zoom_out(self):
        """ Zoom out with - key """
#        x = self.canvas.canvasx(event.x)  # get coordinates of the event on the canvas
#        y = self.canvas.canvasy(event.y)
#        if self.outside(x, y): return  # zoom only inside image area
        scale = 1.0
        if round(self.__min_side * self.imscale) < 30: return  # image is less than 30 pixels
        self.imscale /= self.__delta
        scale        /= self.__delta
        # Take appropriate image from the pyramid
        k = self.imscale * self.__ratio  # temporary coefficient
        self.__curr_img = min((-1) * int(math.log(k, self.__reduction)), len(self.__pyramid) - 1)
        self.__scale = k * math.pow(self.__reduction, max(0, self.__curr_img))
        #
        self.canvas.scale('all', .5, .5, scale, scale)  # rescale all objects
#        print(self.canvas.gettags())
#        print("check2")
        # Redraw some figures before showing image on the screen
        self.redraw_figures()  # method for child classes
        self.__show_image()        

    def __keystroke(self, event):
        """ Scrolling with the keyboard.
            Independent from the language of the keyboard, CapsLock, <Ctrl>+<key>, etc. """
        if event.state - self.__previous_state == 4:  # means that the Control key is pressed
            pass  # do nothing if Control key is pressed
        else:
            self.__previous_state = event.state  # remember the last keystroke state
            # Up, Down, Left, Right keystrokes
            if event.keycode in [68, 39, 102]:  # scroll right, keys 'd' or 'Right'
                self.__scroll_x('scroll',  1, 'unit', event=event)
            elif event.keycode in [65, 37, 100]:  # scroll left, keys 'a' or 'Left'
                self.__scroll_x('scroll', -1, 'unit', event=event)
            elif event.keycode in [87, 38, 104]:  # scroll up, keys 'w' or 'Up'
                self.__scroll_y('scroll', -1, 'unit', event=event)
            elif event.keycode in [83, 40, 98]:  # scroll down, keys 's' or 'Down'
                self.__scroll_y('scroll',  1, 'unit', event=event)

    def crop(self, bbox):
        """ Crop rectangle from the image and return it """
        if self.__huge:  # image is huge and not totally in RAM
            band = bbox[3] - bbox[1]  # width of the tile band
            self.__tile[1][3] = band  # set the tile height
            self.__tile[2] = self.__offset + self.imwidth * bbox[1] * 3  # set offset of the band
            self.__image.close()
            self.__image = Image.open(self.path)  # reopen / reset image
            self.__image.size = (self.imwidth, band)  # set size of the tile band
            self.__image.tile = [self.__tile]
            return self.__image.crop((bbox[0], 0, bbox[2], band))
        else:  # image is totally in RAM
            return self.__pyramid[0].crop(bbox)

    def destroy(self):
        """ ImageFrame destructor """
        self.__image.close()
        map(lambda i: i.close, self.__pyramid)  # close all pyramid images
        del self.__pyramid[:]  # delete pyramid list
        del self.__pyramid  # delete pyramid variable
        self.canvas.destroy()
        self.__imframe.destroy()

#class MainWindow(ttk.Frame):
#    """ Main window class """
#    def __init__(self, mainframe, path):
#        """ Initialize the main Frame """
#        ttk.Frame.__init__(self, master=mainframe)
#        self.master.title('Advanced Zoom v3.0')
#        self.master.geometry('800x600')  # size of the main window
#        self.master.rowconfigure(0, weight=1)  # make the CanvasImage widget expandable
#        self.master.columnconfigure(0, weight=1)
#        canvas = CanvasImage(self.master, path)  # create widget
#        canvas.grid(row=0, column=0)  # show widget

#filename = './data/img_plg5.png'  # place path to your image here
#filename = 'd:/Data/yandex_z18_1-1.tif'  # huge TIFF file 1.4 GB
#filename = 'd:/Data/The_Garden_of_Earthly_Delights_by_Bosch_High_Resolution.jpg'
#filename = 'd:/Data/The_Garden_of_Earthly_Delights_by_Bosch_High_Resolution.tif'
#filename = 'd:/Data/heic1502a.tif'
#filename = 'd:/Data/land_shallow_topo_east.tif'
#filename = 'd:/Data/X1D5_B0002594.3FR'
#app = MainWindow(tk.Tk(), path=filename)
#app.mainloop()













from __future__ import print_function
import os
import tempfile
import contextlib
import shutil


import tkinter as tk
from tkinter import ttk

from tnorm.GUI.canvas_image import CanvasImage


@contextlib.contextmanager
def make_temp_directory():
    temp_dir = tempfile.mkdtemp()
    try:
        yield temp_dir
    finally:
        shutil.rmtree(temp_dir)


def make_canvas_image(tnorm_app, tab, filename):
    tnorm_app.ZoomFrame = tk.Frame(tab)
    tnorm_app.ZoomFrame.pack(side=tk.LEFT,expand=False, fill=tk.Y)
    tnorm_app.ImageFrame = tk.Frame(tab)  # placeholder of the ImageFrame object
    tnorm_app.ImageFrame.pack(side=tk.RIGHT, expand=True, fill=tk.BOTH)

    tnorm_app.ImageFrame.rowconfigure(0, weight=1)  # make the CanvasImage widget expandable
    tnorm_app.ImageFrame.columnconfigure(0, weight=1)
    canvas = CanvasImage(tnorm_app, filename)
    canvas.grid(row=0, column=0)


def make_hasse(tnorm_app, tab):
#    try:
    if tab == tnorm_app.NormBallTab:
        G = tnorm_app.ball.get_hasse_diagram
    elif tab == tnorm_app.DualNormBallTab:
        G = tnorm_app.dual_ball.get_hasse_diagram
    with make_temp_directory() as temp_dir:
        filename = os.path.join(temp_dir, 'hasse.png')
        G.save_image(filename)
        make_canvas_image(tnorm_app, tab, filename)

#    except Exception as e: print('Error: {}'.format(e))
    tnorm_app.stop_spin()


def _show_polyhedron(tnorm_app, online, tab):
    #try:
    if tab == tnorm_app.NormBallTab:
        ball = tnorm_app.ball
        dual = False
    elif tab == tnorm_app.DualNormBallTab:
        ball = tnorm_app.dual_ball
        dual = True
    if not ball.is_compact():
        P = ball.polyhedron_mod_rays()
    else:
        P = ball.polyhedron()
    if P.dim() in [3,4]:
        ball.plot(viewer='x3d', online=online)
    elif P.dim() in [1,2]:
        with make_temp_directory() as temp_dir:
            filename = os.path.join(temp_dir,'plot2d.png')
            plt = ball._plot2d(dual)
            plt.save(filename)
            make_canvas_image(tnorm_app, tab, filename)
    else:
        print('Error: can\'t plot a polyhedron of dimension > 4.')
    #except Exception as e: print('Error: {}'.format(e))
    tnorm_app.stop_spin()











